var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _GossipLogService_instances, _GossipLogService_registrarId, _GossipLogService_pubsub, _GossipLogService_started, _GossipLogService_sync;
import { TypedEventEmitter, TopicValidatorResult, } from "@libp2p/interface";
import { GossipSub } from "@chainsafe/libp2p-gossipsub";
import { logger } from "@libp2p/logger";
import PQueue from "p-queue";
import { anySignal } from "any-signal";
import * as lp from "it-length-prefixed";
import { pipe } from "it-pipe";
import { assert } from "@canvas-js/utils";
import { Event } from "#protocols/events";
import { DEFAULT_PROTOCOL_SELECT_TIMEOUT, MAX_INBOUND_STREAMS, MAX_OUTBOUND_STREAMS, SYNC_RETRY_INTERVAL, SYNC_RETRY_LIMIT, second, } from "../constants.js";
import { decodeId } from "../schema.js";
import { Client, decodeRequests, encodeResponses } from "../sync/index.js";
import { DelayableController, SyncDeadlockError, SyncTimeoutError, wait } from "../utils.js";
import { Server } from "../sync/server.js";
export const getProtocol = (topic) => `/canvas/sync/v1/${topic}`;
export class GossipLogService extends TypedEventEmitter {
    static extractGossipSub(components) {
        const { pubsub } = components;
        assert(pubsub !== undefined, "pubsub service not found");
        assert(pubsub instanceof GossipSub);
        return pubsub;
    }
    constructor(components, messageLog, init) {
        super();
        _GossipLogService_instances.add(this);
        this.components = components;
        this.messageLog = messageLog;
        this.protocol = getProtocol(this.messageLog.topic);
        this.syncQueue = new PQueue({ concurrency: 1 });
        this.syncQueuePeers = new Set();
        this.topologyPeers = new Set();
        this.controller = new AbortController();
        _GossipLogService_registrarId.set(this, null);
        _GossipLogService_pubsub.set(this, void 0);
        _GossipLogService_started.set(this, false);
        this.forwardEvent = (event) => this.safeDispatchEvent(event.type, event);
        this.handleMessage = ({ detail: { msgId, propagationSource, msg } }) => {
            if (msg.type !== "signed" || msg.topic !== this.messageLog.topic) {
                return;
            }
            const sourceId = propagationSource.toString();
            let event;
            try {
                event = Event.decode(msg.data);
            }
            catch (err) {
                this.log.error("error decoding gossipsub message: %O", err);
                __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").reportMessageValidationResult(msgId, sourceId, TopicValidatorResult.Reject);
                return;
            }
            if (event.insert !== undefined) {
                this.handleInsert({ msgId, propagationSource, from: msg.from }, event.insert).then((result) => __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").reportMessageValidationResult(msgId, sourceId, result), (err) => {
                    this.log.error("error handling insert event: %O", err);
                    __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").reportMessageValidationResult(msgId, sourceId, TopicValidatorResult.Reject);
                });
            }
            else if (event.update !== undefined) {
                this.handleUpdate({ msgId, propagationSource, from: msg.from }, event.update).then((result) => __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").reportMessageValidationResult(msgId, sourceId, result), (err) => {
                    this.log.error("error handling update event: %O", err);
                    __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").reportMessageValidationResult(msgId, sourceId, TopicValidatorResult.Reject);
                });
            }
            else {
                this.log.error("error decoding gossipsub message (invalid type)");
                __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").reportMessageValidationResult(msgId, sourceId, TopicValidatorResult.Reject);
            }
        };
        this.handleIncomingStream = async ({ connection, stream }) => {
            const peerId = connection.remotePeer;
            this.log("opened incoming stream %s from peer %p", stream.id, peerId);
            const timeoutController = new DelayableController(3 * second);
            const signal = anySignal([this.controller.signal, timeoutController.signal]);
            signal.addEventListener("abort", (err) => {
                if (stream.status === "open") {
                    stream.abort(new Error("TIMEOUT"));
                }
            });
            try {
                await this.messageLog.serve(async (source) => {
                    const server = new Server(source);
                    await pipe(stream.source, lp.decode, decodeRequests, async function* (reqs) {
                        for await (const req of reqs) {
                            timeoutController.delay();
                            yield req;
                        }
                    }, (reqs) => server.handle(reqs), encodeResponses, lp.encode, stream.sink);
                }, { targetId: peerId.toString() });
                this.log("closed incoming stream %s from peer %p", stream.id, peerId);
            }
            catch (err) {
                if (err instanceof Error && err.message === "TIMEOUT") {
                    this.log.error("timed out incoming stream %s from peer %p", stream.id, peerId);
                    stream.abort(err);
                }
                else if (err instanceof Error) {
                    this.log.error("aborting incoming stream %s from peer %p: %O", stream.id, peerId, err);
                    stream.abort(err);
                }
                else {
                    this.log.error("aborting incoming stream %s from peer %p: %O", stream.id, peerId, err);
                    stream.abort(new Error("internal error"));
                }
            }
            finally {
                signal.clear();
            }
        };
        this.log = logger(`canvas:gossiplog:[${this.messageLog.topic}]:service`);
        __classPrivateFieldSet(this, _GossipLogService_pubsub, GossipLogService.extractGossipSub(components), "f");
        this.maxInboundStreams = init.maxInboundStreams ?? MAX_INBOUND_STREAMS;
        this.maxOutboundStreams = init.maxOutboundStreams ?? MAX_OUTBOUND_STREAMS;
    }
    isStarted() {
        return __classPrivateFieldGet(this, _GossipLogService_started, "f");
    }
    async start() {
        this.log("start");
        this.messageLog.addEventListener("sync", this.forwardEvent);
        this.messageLog.addEventListener("commit", this.forwardEvent);
        this.messageLog.addEventListener("message", this.forwardEvent);
        __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").addEventListener("gossipsub:message", this.handleMessage);
        await this.components.registrar.handle(this.protocol, this.handleIncomingStream, {
            maxInboundStreams: this.maxInboundStreams,
            maxOutboundStreams: this.maxOutboundStreams,
        });
        __classPrivateFieldSet(this, _GossipLogService_registrarId, await this.components.registrar.register(this.protocol, {
            notifyOnTransient: false,
            onConnect: async (peerId, connection) => {
                this.topologyPeers.add(peerId.toString());
                this.log("connected to peer %p", peerId);
                // having one peer wait an initial randomized interval
                // reduces the likelihood of deadlock to near-zero,
                // but it could still happen.
                if (connection.direction === "inbound") {
                    const interval = second + Math.floor(Math.random() * SYNC_RETRY_INTERVAL);
                    this.log("waiting an initial %dms", interval);
                    await this.wait(interval);
                }
                this.scheduleSync(peerId);
            },
            onDisconnect: (peerId) => {
                this.log("disconnected from %p", peerId);
                this.topologyPeers.delete(peerId.toString());
            },
        }), "f");
        __classPrivateFieldSet(this, _GossipLogService_started, true, "f");
    }
    async afterStart() {
        this.log("afterStart");
        __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").subscribe(this.messageLog.topic);
    }
    async beforeStop() {
        this.log("beforeStop");
        __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").unsubscribe(this.messageLog.topic);
    }
    async stop() {
        this.log("stop");
        __classPrivateFieldSet(this, _GossipLogService_started, false, "f");
        __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").removeEventListener("gossipsub:message", this.handleMessage);
        this.messageLog.removeEventListener("sync", this.forwardEvent);
        this.messageLog.removeEventListener("commit", this.forwardEvent);
        this.messageLog.removeEventListener("message", this.forwardEvent);
        await this.components.registrar.unhandle(this.protocol);
        if (__classPrivateFieldGet(this, _GossipLogService_registrarId, "f") !== null) {
            this.components.registrar.unregister(__classPrivateFieldGet(this, _GossipLogService_registrarId, "f"));
            __classPrivateFieldSet(this, _GossipLogService_registrarId, null, "f");
        }
    }
    async append(payload, options = {}) {
        const { id, signature, message } = await this.messageLog.write((txn) => this.messageLog.append(txn, payload, options));
        const [key, value] = this.messageLog.encode(signature, message);
        const data = Event.encode({ insert: { key, value } });
        const recipients = __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").publish(this.messageLog.topic, data).then(({ recipients }) => {
            this.log("published message %s to %d recipients %O", id, recipients.length, recipients);
            return recipients;
        }, (err) => {
            this.log.error("failed to publish event: %O", err);
            return [];
        });
        return { id, signature, message, recipients };
    }
    async insert(signature, message) {
        assert(message.topic === this.messageLog.topic, "wrong topic");
        const { id } = await this.messageLog.write((txn) => this.messageLog.insert(txn, signature, message));
        const [key, value] = this.messageLog.encode(signature, message);
        const data = Event.encode({ insert: { key, value } });
        const recipients = __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").publish(message.topic, data).then(({ recipients }) => {
            this.log("published message %s to %d recipients %O", id, recipients.length, recipients);
            return recipients;
        }, (err) => {
            this.log.error("failed to publish event: %O", err);
            return [];
        });
        return { id, recipients };
    }
    async handleInsert({ msgId, propagationSource, from }, { key, value }) {
        this.log("handling insert event %s from %p (via %p)", msgId, from, propagationSource);
        const [id, signature, message] = this.messageLog.decode(value);
        assert(decodeId(key) === id, "invalid message key");
        await this.messageLog.verifySignature(signature, message);
        const result = await this.messageLog.write(async (txn) => {
            const missingParents = await this.messageLog.getMissingParents(txn, message.parents);
            if (missingParents.size !== 0) {
                this.log("missing %d/%d parents", missingParents.size, message.parents.length);
                return TopicValidatorResult.Ignore;
            }
            try {
                await this.messageLog.apply(txn, id, signature, message, [key, value]);
                return TopicValidatorResult.Accept;
            }
            catch (err) {
                return TopicValidatorResult.Reject;
            }
        });
        // Need to sync
        if (result === TopicValidatorResult.Ignore) {
            this.scheduleSync(propagationSource);
        }
        return result;
    }
    async handleUpdate({ msgId, propagationSource, from }, { heads }) {
        this.log("handling update event %s from %p (via %p)", msgId, from, propagationSource);
        const result = await this.messageLog.read(async (txn) => {
            const missingParents = new Set();
            for (const key of heads) {
                const leaf = await txn.messages.getNode(0, key);
                if (leaf === null) {
                    missingParents.add(decodeId(key));
                }
            }
            if (missingParents.size === 0) {
                return TopicValidatorResult.Accept;
            }
            else {
                return TopicValidatorResult.Ignore;
            }
        });
        // Need to sync
        if (result === TopicValidatorResult.Ignore) {
            this.scheduleSync(propagationSource);
        }
        return result;
    }
    scheduleSync(peerId) {
        const id = peerId.toString();
        if (this.syncQueuePeers.has(id)) {
            this.log("already queued sync with %p", peerId);
            return;
        }
        this.log("scheduling sync with %p", peerId);
        this.syncQueuePeers.add(id);
        this.syncQueue
            .add(() => this.sync(peerId), { priority: 0 })
            .catch((err) => this.log.error("sync failed: %O", err))
            .finally(() => this.syncQueuePeers.delete(id));
    }
    async sync(peerId) {
        this.log("initiating sync with %p", peerId);
        const connection = this.components.connectionManager
            .getConnections(peerId)
            .find((connection) => connection.transient === false);
        if (connection === undefined) {
            this.log("no longer connected to %p", peerId);
            return;
        }
        for (let n = 0; n < SYNC_RETRY_LIMIT; n++) {
            try {
                await __classPrivateFieldGet(this, _GossipLogService_instances, "m", _GossipLogService_sync).call(this, connection);
                return;
            }
            catch (err) {
                if (err instanceof SyncDeadlockError) {
                    this.log("deadlock with %p", peerId);
                }
                else {
                    this.log.error("failed to sync with peer: %O", err);
                }
                if (this.controller.signal.aborted) {
                    break;
                }
                else {
                    const interval = Math.floor(Math.random() * SYNC_RETRY_INTERVAL);
                    this.log("waiting %dms before trying again (%d/%d)", interval, n + 1, SYNC_RETRY_LIMIT);
                    await this.wait(interval);
                    continue;
                }
            }
        }
        throw new Error("exceeded sync retry limit");
    }
    async wait(interval) {
        await wait(interval, { signal: this.controller.signal });
    }
}
_GossipLogService_registrarId = new WeakMap(), _GossipLogService_pubsub = new WeakMap(), _GossipLogService_started = new WeakMap(), _GossipLogService_instances = new WeakSet(), _GossipLogService_sync = async function _GossipLogService_sync(connection) {
    if (connection.status !== "open") {
        throw new Error("connection closed");
    }
    const peerId = connection.remotePeer;
    const sourceId = peerId.toString();
    const selectProtocolSignal = AbortSignal.timeout(DEFAULT_PROTOCOL_SELECT_TIMEOUT);
    const stream = await connection
        .newStream(this.protocol, { negotiateFully: false, signal: selectProtocolSignal })
        .catch((err) => {
        this.log.error("failed to open outgoing stream: %O", err);
        throw err;
    });
    this.log("opened outgoing stream %s to peer %p", stream.id, peerId);
    const timeoutController = new DelayableController(3 * second);
    const signal = anySignal([this.controller.signal, timeoutController.signal]);
    signal.addEventListener("abort", (err) => {
        if (stream.status === "open") {
            stream.abort(new SyncTimeoutError());
        }
    });
    this.log("starting sync with peer %p", peerId);
    let heads = null;
    const client = new Client(stream);
    try {
        heads = await this.messageLog.write(async (txn) => {
            let messageCount = 0;
            for await (const [_, signature, message, entry] of this.messageLog.sync(txn, client, { sourceId })) {
                timeoutController.delay();
                await this.messageLog.insert(txn, signature, message, entry);
                messageCount += 1;
            }
            if (messageCount !== 0) {
                return await txn.getHeads();
            }
            else {
                return null;
            }
        });
    }
    finally {
        signal.clear();
        client.end();
        this.log("closed outgoing stream %s to peer %p", stream.id, peerId);
    }
    if (heads !== null) {
        const data = Event.encode({ update: { heads } });
        __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").publish(this.messageLog.topic, data).then(({ recipients }) => this.log("published update event to %d recipients %O", recipients.length, recipients), (err) => this.log.error("failed to publish update event: %O", err));
    }
};
export const gossiplog = (messageLog, init = {}) => (components) => new GossipLogService(components, messageLog, init);
